package at.ac.tuwien.dbai.staff.dvorak.alternation.runtime;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

/**
 * This class uses tabled evaluation and cycle detection to simulate an alternating program.
 *
 * @author Wolfgang Dvorak
 *
 * @param <GInputtape> Container class for the input variables
 * @param <GWorktape> Container class for the work variables
 */
public class ATMcyclic<GInputtape,GWorktape extends InterfaceWorktape<GInputtape>> implements ATM<GInputtape, GWorktape> {

	private Map<Byte,State<GInputtape,GWorktape>> states;
	private byte startState;
	private GWorktape worktape;
	private Set<ComputationNode<GWorktape>> cycleLeader=new HashSet<ComputationNode<GWorktape>>();
	private Stack<ComputationNode<GWorktape>> predecessors=new Stack<ComputationNode<GWorktape>>();
	private Stack<ComputationNode<GWorktape>> cycleLeaderRanking=new Stack<ComputationNode<GWorktape>>();

	/**
	 * Data structure for the tabled evaluation
	 * the tree stucture saves a little bit more information which is useful in some cases
	 */
	private ComputationTreeCyclic<GWorktape> computationTree;

	/**
	 * constructor
	 * @param states Map with states
	 * @param worktape a (empty) worktape
	 * @param startState stateid from the state the computation should start with
	 */
	public ATMcyclic(Map<Byte,State<GInputtape,GWorktape>> states,GWorktape worktape, byte startState){
		this.states=states;
		this.startState=startState;
		this.worktape=worktape;
	}
	/* (non-Javadoc)
	 * @see at.ac.tuwien.dbai.staff.dvorak.atm.ATM#compute(GInputtape)
	 */
	public boolean compute(GInputtape input){
		byte stateId=startState;

		computationTree=new ComputationTreeCyclic<GWorktape>();
		for(State<GInputtape,GWorktape> state: states.values()){
			state.setInput(input);
		}
		worktape.reset(input);

		ComputationNode<GWorktape> rootNode = new ComputationNode<GWorktape>(stateId,worktape);
		computationTree.setRoot(rootNode);

		return computeNode(rootNode);
	};

	/**
	 * computes the result of a ComputationNode
	 * @param node ComputationNode
	 * @return
	 */
	private Boolean computeNode(ComputationNode<GWorktape> node) {
		//System.out.println(node);
		ComputationNode<GWorktape> node2;

		// search in computationTree (tabled evaluation)
		if (computationTree.containsKey(node)){
			return computationTree.get(node);
		}

		if (predecessors.contains(node)){
			cycleLeader.add(node);
			return null;
		}

		ResultTuple<GWorktape> result;
		GWorktape worktape=(GWorktape)node.worktape.clone();

		result=states.get(node.state).compute(worktape);

		if (result.accept){
			computationTree.put(node,true);//tabled evaluation
			return true;
		}

		if (result.reject){
			computationTree.put(node,false);//tabled evaluation
			return false;
		}

		Boolean hilf=null;
		boolean returnNull=false;

		if (result.forall){
			/*
			 * What to do if there are no successors ? - default behaviour is accepting
			 * if (result.successors.isEmpty() ){return false;}
			 */
			node.setQuantifier(true);
			predecessors.push(node);
			for (ResultTuple.AtmConfiguration<GWorktape> work : result.successors) {
				node2=new ComputationNode<GWorktape>(work.state,work.worktape);
				computationTree.addEdge(node,node2);
				
				hilf=computeNode(node2);

				if(hilf==null){
					returnNull=true;
					computationTree.addNullEdge(node,node2);
				} else {
					if (!hilf){
						computationTree.put(node,false);//tabled evaluation
						predecessors.pop();
						if (cycleLeader.contains(node)){
							force(node,false);
							if(cycleLeader.isEmpty()){
								killSubCycles();
							}
							return false;
						} else {
							return false;
						}
					}
				}
			}

			if (returnNull){
				predecessors.pop();
				if (cycleLeader.contains(node)){
					cycleLeader.remove(node);
					if(cycleLeader.isEmpty()){
						computationTree.put(node,false);//tabled evaluation
						force(node,false);
						killSubCycles();
						return false;
					} else {
						computationTree.put(node,null);//tabled evaluation
						cycleLeaderRanking.push(node);
						return null;
					}

				} else {
					computationTree.put(node,null);//tabled evaluation
					return null;
				}
			}

			computationTree.put(node,true);//tabled evaluation
			predecessors.pop();
			if (cycleLeader.contains(node)){
				cycleLeader.remove(node);
				force(node,true);
				if(cycleLeader.isEmpty()){
					killSubCycles();
				}
				return true;
			} else {
				return true;
			}
		} else {
			node.setQuantifier(false);
			predecessors.push(node);
			for (ResultTuple.AtmConfiguration<GWorktape> work : result.successors) {
				node2=new ComputationNode<GWorktape>(work.state,work.worktape);
				computationTree.addEdge(node,node2);
				
				hilf=computeNode(node2);

				if(hilf==null){
					returnNull=true;
					computationTree.addNullEdge(node,node2);
				} else {
					if (hilf){
						computationTree.put(node,true);//tabled evaluation
						predecessors.pop();
						if (cycleLeader.contains(node)){
							cycleLeader.remove(node);
							force(node,true);
							if(cycleLeader.isEmpty()){
								killSubCycles();
							}
							return true;
						} else {
							return true;
						}
					}
				}

			}

			if (returnNull){
				predecessors.pop();
				if (cycleLeader.contains(node)){
					cycleLeader.remove(node);
					if(cycleLeader.isEmpty()){
						computationTree.put(node,false);//tabled evaluation
						force(node,false);
						killSubCycles();
						return false;
					} else {
						computationTree.put(node,null);//tabled evaluation
						cycleLeaderRanking.push(node);
						return null;
					}

				} else {
					computationTree.put(node,null);//tabled evaluation
					return null;
				}
			}

			computationTree.put(node,false);//tabled evaluation
			predecessors.pop();
			if (cycleLeader.contains(node)){
				cycleLeader.remove(node);
				force(node,false);
				if(cycleLeader.isEmpty()){
					killSubCycles();
				}
				return false;
			} else {
				return false;
			}
		}
	}

	private void force(ComputationNode<GWorktape> node, boolean acceptance) {
		//System.out.print("force: " +node +" acceptance="+acceptance);
		//System.out.println(": "+ computationTree.getParents(node));
		cycleLeader.remove(node);
		cycleLeaderRanking.remove(node);

		for(ComputationNode<GWorktape> parent : computationTree.getParents(node)){
			if (computationTree.get(parent)==null){
				reevaluate(parent, acceptance);
			}			
		}
		computationTree.removeNullEdges(node);
	}
	
	private void killSubCycles() {
		ComputationNode<GWorktape> node;
		//System.out.println("killSubCycles: " + cycleLeaderRanking);
		//System.out.println("               " + computationTree);
		while(! cycleLeaderRanking.isEmpty()){
			node=cycleLeaderRanking.pop();
			if (!(computationTree.get(node)==null)) throw new RuntimeException("error in kill sub cycles");
			computationTree.put(node,false);//tabled evaluation
			force(node,false);
		}

	}
	
	private void reevaluate(ComputationNode<GWorktape> parent, boolean acceptance) {
		//System.out.print("reevaluate: "+ parent);
		boolean returnNull=false;

		if (parent.isAllQuantified()){
			if (acceptance==false){
				computationTree.put(parent,false);//tabled evaluation
				force(parent,false);
			}else {
				for(ComputationNode<GWorktape> child: computationTree.getChildren(parent)){
					if(computationTree.get(child)==null){						
						returnNull=true;
						break;
					}
					if(computationTree.get(child).equals(false)){						
						throw new RuntimeException("error in reevaluate");
					}
				}
				if (!returnNull){
					computationTree.put(parent,true);
					force(parent,true);
				}
			}
		} else {
			if (acceptance==true){
				computationTree.put(parent,true);//tabled evaluation
				force(parent,true);
			}else {
				for(ComputationNode<GWorktape> child: computationTree.getChildren(parent)){
					if(computationTree.get(child)==null){
						returnNull=true;
						break;
					}
					if(computationTree.get(child).equals(true)){						
						throw new RuntimeException("error in reevaluate");
					}
				}
				if (!returnNull){
					computationTree.put(parent,false);
					force(parent,false);
				}
			}
		}
	}
	
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.dbai.staff.dvorak.atm.ATM#getComputationTree()
	 */
	public ComputationTree<GWorktape> getComputationTree(){
		return this.computationTree;
	}
}